import UIKit
import Mapbox

class ViewController: UIViewController, MGLMapViewDelegate {

    @IBOutlet var map: MGLMapView!
    @IBOutlet var buttonStack: UIStackView!

    let actions = [
        "Color Buildings By Decade",
        "Highlight Undated Buildings",
        "Query Individual Buildings",
        "Examine Sibling Buildings",
        "Play Timeline Animation",
    ]

    let customBuildingsMapID = "justin.0079auf9"

    let prefix = "pdx-buildings"

    let startingCenter = CLLocationCoordinate2D(latitude: 45.52, longitude: -122.681944)
    let startingZoom = 9.0

    let lowYear = 1840
    let highYear = 2010
    let thisYear = Calendar.current.component(.year, from: Date())

    let taggedDefaultColor = UIColor.lightGray
    let taggedDefaultOpacity = 0.25

    let untaggedDefaultColor = UIColor.darkGray
    let untaggedDefaultOpacity = 0.25

    let taggedOpacity = 0.75

    let untaggedHighlightColor = UIColor.yellow

    let popup = UILabel(frame: CGRect(x: 0, y: 0, width: 50, height: 30))

    override func viewDidLoad() {
        super.viewDidLoad()

        title = "Portland's Buildings"

        navigationController?.navigationBar.barStyle = .black
        navigationController?.navigationBar.tintColor = UIColor.lightGray
        navigationController?.navigationBar.titleTextAttributes = [NSForegroundColorAttributeName: UIColor.lightGray]
        map.tintColor = UIColor.lightGray

        navigationItem.rightBarButtonItem = UIBarButtonItem(image: UIImage(named: "gears"),
                                                            style: .plain,
                                                            target: self,
                                                            action: #selector(tappedSettings))

        map.styleURL = MGLStyle.darkStyleURL(withVersion: MGLStyleDefaultVersion)
        map.setCenter(startingCenter, zoomLevel: startingZoom, animated: false)

        for title in actions {
            let button = TintedButton(type: .custom)
            button.addTarget(self, action: #selector(tappedAction(_:)), for: .touchUpInside)
            button.bounds = CGRect(x: 0, y: 0, width: 200, height: 50)
            button.setTitle(title, for: .normal)
            button.setTitleColor(.black, for: .normal)
            button.setTitleColor(.white, for: .selected)
            button.titleLabel?.font = UIFont.systemFont(ofSize: 13)
            button.backgroundColor = UIColor.white.withAlphaComponent(0.75)
            button.layer.borderColor = UIColor.darkGray.cgColor
            button.layer.borderWidth = 1
            button.layer.cornerRadius = 5
            buttonStack.addArrangedSubview(button)
        }
        buttonStack.alpha = 0
    }

    func mapView(_ mapView: MGLMapView, didFinishLoading style: MGLStyle) {
        setupPopup()
        hideDefaultBuildings()
        addCustomBuildingsSource()
        setupTaggedBuildings()
        setupUntaggedBuildings()
    }

    func setupPopup() {
        popup.backgroundColor = UIColor.white.withAlphaComponent(0.9)
        popup.font = UIFont.systemFont(ofSize: 12)
        popup.textAlignment = .center
        popup.alpha = 0
        view.addSubview(popup)
    }

    func tappedSettings() {
        UIView.animate(withDuration: 0.25) { [unowned self] in
            self.buttonStack.alpha = (self.buttonStack.alpha == 1 ? 0 : 1)
        }
    }

    func tappedAction(_ button: UIButton) {
        UIView.animate(withDuration: 0.25) { [unowned self] in
            self.buttonStack.alpha = 0
        }

        button.isSelected = !button.isSelected

        switch actions.index(of: button.title(for: .normal)!)! {
        case 0:
            // toggle buildings with build dates ("tagged")
            toggleTaggedBuildings(isEnabled: button.isSelected)
        case 1:
            // toggle buildings with no build date data ("untagged")
            toggleUntaggedBuildings(isEnabled: button.isSelected)
        case 2:
            // toggle query mode for interactively browsing dates
            removeQueryFunctionality()
            if button.isSelected {
                map.isUserInteractionEnabled = false
                view.addGestureRecognizer(UIPanGestureRecognizer(target: self, action: #selector(handleQueryPan(_:))))
                if let defaultBuildings = map.style.layer(withIdentifier: "building") {
                    // create & add a custom source & fill layer for highlighting the query
                    let highlightSource = MGLShapeSource(identifier: "\(prefix)-query", shape: nil, options: nil)
                    map.style.addSource(highlightSource)
                    let highlightLayer = MGLFillStyleLayer(identifier: "\(prefix)-query", source: highlightSource)
                    highlightLayer.fillColor = MGLStyleValue(rawValue: .yellow)
                    map.style.insertLayer(highlightLayer, below: defaultBuildings)
                }
            }
            // disable sibling mode (mutually exclusive)
            (buttonStack.arrangedSubviews[3] as! UIButton).isSelected = false
        case 3:
            // toggle "sibling mode" for highlighting buildings with the same build date
            removeQueryFunctionality()
            if button.isSelected {
                map.isUserInteractionEnabled = false
                view.addGestureRecognizer(UIPanGestureRecognizer(target: self, action: #selector(handleSiblingPan(_:))))
            }
            // disable query mode (mutually exclusive)
            (buttonStack.arrangedSubviews[2] as! UIButton).isSelected = false
        case 4:
            // play a timeline animation showing Portland being built
            if button.isSelected {
                toggleTaggedBuildings(isEnabled: false)
                toggleUntaggedBuildings(isEnabled: false)
                if let tagged = map.style.layer(withIdentifier: "\(prefix)-tagged") as? MGLFillStyleLayer {
                    tagged.isVisible = false
                }
                if let untagged = map.style.layer(withIdentifier: "\(prefix)-untagged") as? MGLFillStyleLayer {
                    untagged.isVisible = false
                }
                DispatchQueue.main.asyncAfter(deadline: .now() + .seconds(1)) { [unowned self] in
                    self.startTimelineAnimation()
                }
            }
        default:
            break
        }
    }

    func removeQueryFunctionality() {
        map.isUserInteractionEnabled = true

        view.gestureRecognizers?.forEach { (gesture) in
            view.removeGestureRecognizer(gesture)
        }

        if let highlightSource = map.style.source(withIdentifier: "\(prefix)-query"),
           let highlightLayer = map.style.layer(withIdentifier: "\(prefix)-query") {
            map.style.removeSource(highlightSource)
            map.style.removeLayer(highlightLayer)
        }

        if let siblings = map.style.layer(withIdentifier: "\(prefix)-siblings") {
            map.style.removeLayer(siblings)
        }
    }

    func handleQueryPan(_ pan: UIPanGestureRecognizer) {
        var hidden = false
        if pan.state == .began || pan.state == .changed {
            let point = pan.location(in: pan.view)
            // query not only the normal tagged layer but also any decade layers
            var layers = Set<String>()
            layers.insert("\(prefix)-tagged")
            for decade in stride(from: lowYear, through: highYear, by: 10) {
                layers.insert("\(prefix)-tagged-\(decade)")
            }
            let features = map.visibleFeatures(in: CGRect(x: point.x - 10, y: point.y - 10, width: 20, height: 20), styleLayerIdentifiers: layers)
            if features.count > 0,
               let feature = features.first,
               let year = feature.attribute(forKey: "YEAR_BUILT") as? NSNumber,
               let highlightSource = map.style.source(withIdentifier: "\(prefix)-query") as? MGLShapeSource,
               let shape = feature as? MGLShape {
                // update the query in the highlight source
                highlightSource.shape = shape
                popup.center = CGPoint(x: pan.location(in: pan.view).x, y: pan.location(in: pan.view).y - 50)
                popup.text = "\(year)"
                hidden = false
            } else {
                hidden = true
            }
        } else if pan.state == .ended || pan.state == .cancelled {
            if let highlightSource = map.style.source(withIdentifier: "\(prefix)-query") as? MGLShapeSource {
                // effectively hide the query highlight
                highlightSource.shape = nil
            }
            hidden = true
        }
        UIView.animate(withDuration: 0.25) { [unowned self] in
            self.popup.alpha = (hidden ? 0 : 1)
        }
    }

    func handleSiblingPan(_ pan: UIPanGestureRecognizer) {
        var hidden = false
        if pan.state == .began || pan.state == .changed {
            let point = pan.location(in: pan.view)
            // query both normal tagged & any decade visualization layers
            var layers = Set<String>()
            layers.insert("\(prefix)-tagged")
            for decade in stride(from: lowYear, through: highYear, by: 10) {
                layers.insert("\(prefix)-tagged-\(decade)")
            }
            let features = map.visibleFeatures(in: CGRect(x: point.x - 10, y: point.y - 10, width: 20, height: 20), styleLayerIdentifiers: layers)
            if features.count > 0 {
                let year = features.first!.attribute(forKey: "YEAR_BUILT")! as! NSNumber
                if map.style.layer(withIdentifier: "\(prefix)-siblings") == nil,
                   let defaultBuildings = map.style.layer(withIdentifier: "building"),
                   let source = map.style.source(withIdentifier: "\(prefix)-buildings") {
                    // add a custom fill layer for sibling buildings (but only the first time needed)
                    let siblings = MGLFillStyleLayer(identifier: "\(prefix)-siblings", source: source)
                    siblings.sourceLayerIdentifier = "buildings"
                    siblings.fillColor = MGLStyleValue(rawValue: .green)
                    siblings.fillOpacity = MGLStyleValue(rawValue: NSNumber(value: 0.75))
                    map.style.insertLayer(siblings, below: defaultBuildings)
                }
                if let siblings = map.style.layer(withIdentifier: "\(prefix)-siblings") as? MGLFillStyleLayer {
                    // update the filter to only style sibling buildings for the query year
                    siblings.predicate = NSPredicate(format: "YEAR_BUILT == %@", year)
                }
                popup.center = CGPoint(x: point.x, y: point.y - 50)
                popup.text = "\(year)"
                hidden = false
            }
        } else if pan.state == .ended || pan.state == .cancelled {
            if let siblings = map.style.layer(withIdentifier: "\(prefix)-siblings") as? MGLFillStyleLayer {
                map.style.removeLayer(siblings)
            }
            hidden = true
        }
        UIView.animate(withDuration: 0.25) { [unowned self] in
            self.popup.alpha = (hidden ? 0 : 1)
        }
    }

    func hideDefaultBuildings() {
        // hide the default style's OpenStreetMap-based buildings layer
        if let defaultBuildings = map.style.layer(withIdentifier: "building") {
            defaultBuildings.isVisible = false
        }
    }

    func addCustomBuildingsSource() {
        // add custom Shapefile → GeoJSON → Vector Tiles → Mapbox-hosted buildings data
        if map.style.sources.contains("\(prefix)-buildings") == false {
            if let sourceURL = URL(string: "mapbox://\(customBuildingsMapID)") {
                let source = MGLVectorSource(identifier: "\(prefix)-buildings", configurationURL: sourceURL)
                map.style.addSource(source)
            }
        }
    }

    func setupTaggedBuildings() {
        // "tagged" buildings have valid build dates in the source data
        if map.style.layers.contains("\(prefix)-tagged") == false {
            if let defaultBuildings = map.style.layer(withIdentifier: "building"),
               let source = map.style.source(withIdentifier: "\(prefix)-buildings") {
                let tagged = MGLFillStyleLayer(identifier: "\(prefix)-tagged", source: source)
                tagged.sourceLayerIdentifier = "buildings"
                tagged.predicate = NSPredicate(format: "YEAR_BUILT != 0 AND YEAR_BUILT != null")
                tagged.fillColor = MGLStyleValue(rawValue: taggedDefaultColor)
                tagged.fillOpacity = MGLStyleValue(rawValue: NSNumber(value: taggedDefaultOpacity))
                map.style.insertLayer(tagged, below: defaultBuildings)
            }
        }
    }

    func setupUntaggedBuildings() {
        // "untagged" buildings have invalid build dates in the source data
        if map.style.layers.contains("\(prefix)-untagged") == false {
            if let defaultBuildings = map.style.layer(withIdentifier: "building"),
               let source = map.style.source(withIdentifier: "\(prefix)-buildings") {
                let untagged = MGLFillStyleLayer(identifier: "\(prefix)-untagged", source: source)
                untagged.sourceLayerIdentifier = "buildings"
                untagged.predicate = NSPredicate(format: "YEAR_BUILT == 0 OR YEAR_BUILT == null")
                untagged.fillColor = MGLStyleValue(rawValue: untaggedDefaultColor)
                untagged.fillOpacity = MGLStyleValue(rawValue: NSNumber(value: untaggedDefaultOpacity))
                map.style.insertLayer(untagged, below: defaultBuildings)
            }
        }
    }

    func toggleTaggedBuildings(isEnabled: Bool) {
        if let tagged = map.style.layer(withIdentifier: "\(prefix)-tagged"),
           let source = map.style.source(withIdentifier: "\(prefix)-buildings"),
           let defaultBuildings = map.style.layer(withIdentifier: "building") {
            tagged.isVisible = !isEnabled
            if isEnabled {
                // work around for no data-driven styling: add a style layer for each decade
                for decade in stride(from: lowYear, through: highYear, by: 10) {
                    let factor = CGFloat(1 - (Double(thisYear - decade) / Double(thisYear - lowYear + 10)))
                    let red = CGFloat(255) * factor
                    let blue = CGFloat(255 * (1 - factor))
                    let color = UIColor(red: red / 255, green: 0, blue: blue / 255, alpha: 1)
                    let layer = MGLFillStyleLayer(identifier: "\(prefix)-tagged-\(decade)", source: source)
                    layer.predicate = NSPredicate(format: "YEAR_BUILT >= %@ AND YEAR_BUILT <= %@", NSNumber(value: decade), NSNumber(value: decade + 9))
                    layer.sourceLayerIdentifier = "buildings"
                    layer.fillColor = MGLStyleValue(rawValue: color)
                    layer.fillOpacity = MGLStyleValue(rawValue: NSNumber(value: taggedOpacity))
                    map.style.insertLayer(layer, below: defaultBuildings)
                }
            } else {
                for decade in stride(from: lowYear, through: highYear, by: 10) {
                    if let layer = map.style.layer(withIdentifier: "\(prefix)-tagged-\(decade)") {
                        map.style.removeLayer(layer)
                    }
                }
            }
        }
    }

    func toggleUntaggedBuildings(isEnabled: Bool) {
        if let layer = map.style.layer(withIdentifier: "\(prefix)-untagged") as? MGLFillStyleLayer {
            if isEnabled {
                layer.fillColor = MGLStyleValue(rawValue: untaggedHighlightColor)
            } else {
                layer.fillColor = MGLStyleValue(rawValue: untaggedDefaultColor)
            }
        }
    }

    func startTimelineAnimation() {
        navigationItem.rightBarButtonItem?.isEnabled = false
        if let source = map.style.source(withIdentifier: "\(prefix)-buildings"),
           let defaultBuildings = map.style.layer(withIdentifier: "building") {
            let label = UILabel(frame: CGRect(x: 0, y: 0, width: 90, height: 40))
            label.translatesAutoresizingMaskIntoConstraints = false
            label.clipsToBounds = true
            label.layer.cornerRadius = 3
            label.layer.borderColor = UIColor.darkGray.cgColor
            label.layer.borderWidth = 1
            label.backgroundColor = UIColor.white.withAlphaComponent(0.75)
            label.textAlignment = .center
            label.font = UIFont.systemFont(ofSize: 32)
            label.text = "\(self.lowYear)"
            label.tag = 100
            view.addSubview(label)
            view.addConstraints([
                NSLayoutConstraint(item: label,
                                   attribute: .leading,
                                   relatedBy: .equal,
                                   toItem: view,
                                   attribute: .leading,
                                   multiplier: 1,
                                   constant: 20),
                NSLayoutConstraint(item: label,
                                   attribute: .top,
                                   relatedBy: .equal,
                                   toItem: topLayoutGuide,
                                   attribute: .bottom,
                                   multiplier: 1,
                                   constant: 8),
                NSLayoutConstraint(item: label,
                                   attribute: .width,
                                   relatedBy: .equal,
                                   toItem: nil,
                                   attribute: .notAnAttribute,
                                   multiplier: 1,
                                   constant: label.bounds.size.width),
                NSLayoutConstraint(item: label,
                                   attribute: .height,
                                   relatedBy: .equal,
                                   toItem: nil,
                                   attribute: .notAnAttribute,
                                   multiplier: 1,
                                   constant: label.bounds.size.height)])
            buttonStack.isHidden = true
            let highlightLayer = MGLFillStyleLayer(identifier: "\(prefix)-timeline-highlight", source: source)
            highlightLayer.predicate = NSPredicate(value: false)
            highlightLayer.sourceLayerIdentifier = "buildings"
            highlightLayer.fillColor = MGLStyleValue(rawValue: .white)
            highlightLayer.fillOpacity = MGLStyleValue(rawValue: NSNumber(value: 0.75))
            self.map.style.insertLayer(highlightLayer, below: defaultBuildings)
            let completedLayer = MGLFillStyleLayer(identifier: "\(prefix)-timeline-completed", source: source)
            completedLayer.predicate = NSPredicate(value: false)
            completedLayer.sourceLayerIdentifier = "buildings"
            completedLayer.fillColor = MGLStyleValue(rawValue: .red)
            completedLayer.fillOpacity = MGLStyleValue(rawValue: NSNumber(value: 0.75))
            map.style.insertLayer(completedLayer, below: defaultBuildings)
            DispatchQueue.global(qos: .default).async { [unowned self] in
                for year in stride(from: self.lowYear, through: self.highYear, by: 1) {
                    DispatchQueue.main.async { [unowned self] in
                        label.text = "\(year)"
                        if let layer = self.map.style.layer(withIdentifier: "\(self.prefix)-timeline-highlight") as? MGLFillStyleLayer {
                            layer.predicate = NSPredicate(format: "YEAR_BUILT == %@", NSNumber(value: year))
                        }
                    }
                    usleep(100000)
                    DispatchQueue.main.async { [unowned self] in
                        if let layer = self.map.style.layer(withIdentifier: "\(self.prefix)-timeline-completed") as? MGLFillStyleLayer {
                            layer.predicate = NSPredicate(format: "YEAR_BUILT >= %@ AND YEAR_BUILT <= %@", NSNumber(value: self.lowYear), NSNumber(value: year))
                        }
                    }
                }
                sleep(3)
                DispatchQueue.main.async { [unowned self] in
                    self.cleanupTimelineAnimation()
                }
            }
        }
    }

    func cleanupTimelineAnimation() {
        toggleTaggedBuildings(isEnabled: false)
        (buttonStack.arrangedSubviews.last as! UIButton).isSelected = false
        navigationItem.rightBarButtonItem?.isEnabled = true
        if let highlightLayer = map.style.layer(withIdentifier: "\(prefix)-timeline-highlight"),
           let completedLayer = map.style.layer(withIdentifier: "\(prefix)-timeline-completed"),
           let label = view.viewWithTag(100) {
            map.style.removeLayer(highlightLayer)
            map.style.removeLayer(completedLayer)
            buttonStack.isHidden = false
            label.removeFromSuperview()
            toggleTaggedBuildings(isEnabled: false)
        }
    }

}

class TintedButton: UIButton {

    override var isSelected: Bool {
        didSet {
            backgroundColor = (isSelected ? UIColor.darkGray : UIColor.white).withAlphaComponent(0.75)
        }
    }

}
